package net.davidc.egp.snaxe;

import java.awt.*;
import java.io.IOException;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.davidc.egp.common.events.EventManager;
import net.davidc.egp.common.resources.ResourcesException;
import net.davidc.egp.common.slick2d.ColorUtilSlick;
import net.davidc.egp.common.slick2d.SlickGraphicUtil;
import net.davidc.egp.common.slick2d.resources.SlickResources;
import net.davidc.egp.snaxe.events.ChangeSlickStateEvent;
import net.davidc.egp.snaxe.events.PowerUpRiseEvent;
import net.davidc.egp.snaxe.events.StartGameEvent;
import net.davidc.egp.snaxe.networking.Protocol;
import org.newdawn.slick.*;
import org.newdawn.slick.Color;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.Image;
import org.newdawn.slick.state.BasicGameState;
import org.newdawn.slick.state.StateBasedGame;

/**
 * @author David Croft (david@davidc.net)
 * @version $Id: PlayState.java 23 2011-09-18 12:57:34Z david $
 */
public class PlayState extends BasicGameState
{
  private static final Logger log = Logger.getLogger(PlayState.class.getName());

  public static final int STATE_ID = 2;

  private static final int BOARD_WIDTH = 23;
  private static final int BOARD_HEIGHT = 23;

  private static final int SQUARE_WIDTH = 30;
  private static final int SQUARE_HEIGHT = 30;

  private static final int BOARD_X = 260;
  private static final int BOARD_Y = 40;
  private static final int BOARD_WIDTH_PIXELS = BOARD_WIDTH * SQUARE_WIDTH;
  private static final int BOARD_HEIGHT_PIXELS = BOARD_HEIGHT * SQUARE_HEIGHT;

  private static final int SPEEDUP_CONSIDER_FIRST = 20000; // first consider speed up at 20 seconds
  private static final int SPEEDUP_CONSIDER_EVERY = 8000; // then consider every 8 seconds
  private static final int INITIAL_SPEED = 500;
  private static final double SPEEDUP_SPEED = 0.95;

  private static final int INITIAL_COUNTDOWN_SECONDS = 6;

  private static final int QUEUE_BOTTOM_Y = (BOARD_Y + BOARD_HEIGHT_PIXELS); //750;
  private static final int QUEUE_LEFT_X = 15;

  private static final Color[] SNAKE_COLORS = new Color[]{
//          new Color(0xff0000),
//          new Color(0x00ff00),
//          new Color(0x0000ff),
//          new Color(0xffff00),
//          new Color(0xff00ff),
//          new Color(0x00ffff),
//          new Color(0x000000),
//          new Color(0xffffff),
          new Color(0xdf0b01),
          new Color(0x34bd03),
          new Color(0x054fff),
          new Color(0xa70bff),
          new Color(0xf4d400),
          new Color(0x5d180b),
          new Color(0xd47d02),
          new Color(0x00c7f4),
  };

  private static final Point[][] START_POSITIONS = new Point[][]{
          new Point[]{new Point(8, 11), new Point(14, 11)},
          new Point[]{new Point(11, 8), new Point(8, 13), new Point(14, 13)},
          new Point[]{new Point(9, 8), new Point(13, 8), new Point(9, 14), new Point(13, 14)},
          new Point[]{new Point(11, 8), new Point(8, 11), new Point(14, 11), new Point(9, 15), new Point(13, 15)},
          new Point[]{new Point(9, 8), new Point(13, 8), new Point(8, 11), new Point(14, 11), new Point(9, 14), new Point(13, 14)},
          new Point[]{new Point(9, 7), new Point(13, 7), new Point(8, 10), new Point(14, 10), new Point(8, 13), new Point(14, 13), new Point(11, 15)},
          new Point[]{new Point(10, 8), new Point(12, 8), new Point(8, 10), new Point(14, 10), new Point(8, 12), new Point(14, 12), new Point(10, 14), new Point(12, 14)},
  };

  private static final Direction[][] START_FACING = new Direction[][]{
          new Direction[]{Direction.WEST, Direction.EAST},
          new Direction[]{Direction.NORTH, Direction.WEST, Direction.EAST},
          new Direction[]{Direction.WEST, Direction.EAST, Direction.WEST, Direction.EAST},
          new Direction[]{Direction.NORTH, Direction.WEST, Direction.EAST, Direction.WEST, Direction.EAST},
          new Direction[]{Direction.NORTH, Direction.NORTH, Direction.WEST, Direction.EAST, Direction.SOUTH, Direction.SOUTH},
          new Direction[]{Direction.NORTH, Direction.NORTH, Direction.WEST, Direction.EAST, Direction.WEST, Direction.EAST, Direction.SOUTH},
          new Direction[]{Direction.NORTH, Direction.NORTH, Direction.WEST, Direction.EAST, Direction.WEST, Direction.EAST, Direction.SOUTH, Direction.SOUTH}
  };

//  private static final Color PLAYFIELD_BACKGROUND_COLOR = new Color(148, 156, 57);
  private static final Color PLAYFIELD_BACKGROUND_COLOR = new Color(0xa8caca);
  private final static Color SPEED_COLOR = Color.orange;
  private final static Color INVULNERABILITY_COLOR = Color.orange;
  private final static Color RABBIT_COLOR = new Color(0x722500);
  private final static Color SKULL_COLOR = Color.white;

  GameBoard board;
  private SlickResources resources;

  long curTime;
  long lastMove;

  EventManager<Long, PlayState> eventManager;

  public AngelCodeFont powerUpFont;
  public Graphics graphics;
  private float timeBetweenMoves;
  private int nextConsiderSpeedUp;

  int state = 0; // 0 = countdown, 1 = playing, 2 = won

  private Map<Player, Snake> currentGamePlayers;
  private SnaxeGame game;
  private boolean doneHalfMove;

  @Override
  public int getID()
  {
    return STATE_ID;
  }

  public void init(GameContainer container, StateBasedGame game) throws SlickException
  {
    resources = new SlickResources(((SnaxeGame) game).getGlobalResources());
    try {
      resources.loadFromClasspath("/resources.xml", "game");
    }
    catch (ResourcesException e) {
      throw new SlickException("Unable to load resources", e);
    }

    log.info("Initialisation PlayState");

    eventManager = new EventManager<Long, PlayState>();

    powerUpFont = resources.getFont("FONT_NOKIA16");
  }

  @Override
  public void enter(GameContainer container, StateBasedGame game) throws SlickException
  {
    this.game = (SnaxeGame) game;

    board = new GameBoard(BOARD_WIDTH, BOARD_HEIGHT, this);

    currentGamePlayers = new HashMap<Player, Snake>();

    // Count number of players for this next game:

    int numPlayers = 0;
    for (Player player : ((SnaxeGame) game).getSortedPlayers()) {
      if (player.getQueueJoinTime() > 0) {
        numPlayers++;
      }
    }
    if (numPlayers > SnaxeGame.MAX_PLAYERS) numPlayers = SnaxeGame.MAX_PLAYERS;


    int playerNum = 0;
    for (Player player : ((SnaxeGame) game).getSortedPlayers()) {
      if (player.getQueueJoinTime() > 0) {
        addPlayer(player, playerNum, numPlayers);

        playerNum++;
        if (playerNum == SnaxeGame.MAX_PLAYERS) break;
      }
    }

    // For testing, fixed number of players:
//    numPlayers = SnaxeGame.MAX_PLAYERS;
//
//    for (int i = 0; i < SnaxeGame.MAX_PLAYERS; ++i) {
//      addPlayer(((SnaxeGame) game).getSortedPlayers().iterator().next(), i, numPlayers);
//    }


    changeState(0);
  }

  public void changeState(int newStateID)
  {
    curTime = 0;

    eventManager.removeAllEvents();

    if (newStateID == 0) {
      eventManager.addEvent(new StartGameEvent(curTime + (INITIAL_COUNTDOWN_SECONDS * 1000)));
    }
    else if (newStateID == 1) {
      lastMove = 0;

      timeBetweenMoves = INITIAL_SPEED;

      // Beginning at time 30, increase speed every 10 seconds, as long as one player is ahead
      nextConsiderSpeedUp = SPEEDUP_CONSIDER_FIRST;

      board.initRabbits();
    }
    else if (newStateID == 2) {
      eventManager.addEvent(new ChangeSlickStateEvent(curTime + 5000, game, InsertCoinState.STATE_ID));
    }

    state = newStateID;
  }

  public void update(GameContainer container, StateBasedGame game, int delta) throws SlickException
  {
    curTime += delta;

    eventManager.performEvents(curTime, this);
    eventManager.performUpdate(curTime, this);

    if (state == 1) {
      updateState1(delta);
    }
  }

  private void updateState1(int delta)
  {
    // Consider whether to speed the game up

    if (curTime > nextConsiderSpeedUp) {
      nextConsiderSpeedUp += SPEEDUP_CONSIDER_EVERY;
      // Is there a player in the lead?
      // TODO this should require them to be MORE than 20 points ahead
      int highScore = 0;
      int numWithHighScore = 0;
      for (Snake snake : board.snakes) { // automatically excludes dead snakes
        if (snake.score > highScore) {
          highScore = snake.score;
          numWithHighScore = 1;
        }
        else if (snake.score == highScore) {
          numWithHighScore++;
        }
      }

      if (numWithHighScore < 2) {
        timeBetweenMoves *= SPEEDUP_SPEED;
      }
    }


    if (curTime > lastMove + timeBetweenMoves) {
      board.move(curTime);
      lastMove = curTime;
      doneHalfMove = false;
    }

    if (!doneHalfMove && curTime > lastMove + (timeBetweenMoves / 2)) {
      board.halfMove();
      doneHalfMove = true;
    }

    // Is there only one snake left alive?

    if (board.snakes.size() < 2) {
      gameOver();
    }
  }

  private void gameOver()
  {
    changeState(2);

    // send gameover message to all players
    for (Player player : currentGamePlayers.keySet()) {
      try {
        player.sendMessage(Protocol.MESSAGE_OUT_GAMEOVER);
      }
      catch (IOException e) {
        log.log(Level.WARNING, "Unable to send gameover message to " + player);
      }
    }
  }

  public void render(GameContainer container, StateBasedGame game, Graphics g) throws SlickException
  {
//    g.setBackground(new Color(148, 156, 57));
//    g.fillRect(0, 0, container.getWidth(), container.getHeight());
    g.clear();

    g.setFont(resources.getFont("FONT_NOKIA16"));

    g.setColor(PLAYFIELD_BACKGROUND_COLOR);
    g.fillRect(BOARD_X, BOARD_Y, BOARD_WIDTH_PIXELS, BOARD_HEIGHT * SQUARE_HEIGHT);

    g.setColor(Color.white);
    g.setLineWidth(2);
    g.drawRect(BOARD_X, BOARD_Y, BOARD_WIDTH_PIXELS, BOARD_HEIGHT * SQUARE_HEIGHT);

    if (board == null) return;

    for (Snake snake : board.snakes) {
      Color snakeColor = snake.color;
      if (snake.invulnerabilityTimeRemaining > 0) {
        if (snake.invulnerabilityTimeRemaining > 8000 ||
                (snake.invulnerabilityTimeRemaining >= 6000 && curTime % 1000 < 500) ||
                (snake.invulnerabilityTimeRemaining >= 3000 && snake.invulnerabilityTimeRemaining < 6000 && curTime % 700 < 350) ||
                (snake.invulnerabilityTimeRemaining < 3000 && curTime % 400 < 200)) {
          snakeColor = new Color(snakeColor.r, snakeColor.g, snakeColor.b, 0.5f);
        }
      }

      SnakePart part = snake.head;
      while (part != null) {
        Image im = getImageForPart(part);
        if (im == null) {
          log.log(Level.WARNING, "Unable to get image for part " + part);
        }
        else {
          g.drawImage(im, BOARD_X + (part.x * SQUARE_WIDTH), BOARD_Y + (part.y * SQUARE_HEIGHT), snakeColor);
          // TODO remove: Temp: scale it up
//          int x = BOARD_X + (part.x * SQUARE_WIDTH);
//          int y = BOARD_Y + (part.y * SQUARE_HEIGHT);
//          g.drawImage(im, x, y, x + SQUARE_WIDTH - 1, y + SQUARE_HEIGHT - 1,
//                  0, 0, im.getWidth() - 1, im.getHeight() - 1,
//                  snakeColor);
        }
        part = part.next;
      }
    }

    for (Piece piece : board.pieces) {
      Image im = null;
      Color color = null;
      switch (piece.type) {
        case RABBIT:
          im = resources.getImage("IMAGE_RABBIT");
          color = RABBIT_COLOR;
          break;
        case SKULL:
          im = resources.getImage("IMAGE_SKULL");
          color = SKULL_COLOR;
          break;
        case INVULNERABILITY:
          im = resources.getImage("IMAGE_INVULNERABILITY");
          color = INVULNERABILITY_COLOR;
          break;
        case SPEED:
          im = resources.getImage("IMAGE_SPEED");
          color = SPEED_COLOR;
          break;
      }
      if (im == null) {
        log.log(Level.WARNING, "Unable to get image for piece " + piece.type);
      }
      else {
        g.drawImage(im, BOARD_X + (piece.x * SQUARE_WIDTH), BOARD_Y + (piece.y * SQUARE_HEIGHT), color);
        // TODO remove: Temp: scale it up
//        int x = BOARD_X + (piece.x * SQUARE_WIDTH);
//        int y = BOARD_Y + (piece.y * SQUARE_HEIGHT);
//        g.drawImage(im, x, y, x + SQUARE_WIDTH - 1, y + SQUARE_HEIGHT - 1,
//                0, 0, im.getWidth() - 1, im.getHeight() - 1, color);

      }
    }

    this.graphics = g;
    eventManager.performRender(curTime, this);

    // If in countdown, draw countdown
    if (state == 0) {
      AngelCodeFont font = resources.getFont("FONT_NOKIA128NUMGO");
      g.setFont(font);

      int countdown = INITIAL_COUNTDOWN_SECONDS - (int) (curTime / 1000) - 1;
      float proportionThroughThisSecond = (curTime % 1000) / 1000.0f;

      Color color;
      if (proportionThroughThisSecond < 0.5) {
        color = new Color(0, 0, 255, 255);
      }
      else {
        color = new Color(0f, 0f, 1.0f, 1.0f - ((proportionThroughThisSecond - 0.5f) * 2f));
      }
      g.setColor(color);

      String msg;
      if (countdown == 0) msg = "GO!";
      else msg = "" + countdown;
//      drawStringScaled(g, BOARD_X + (BOARD_WIDTH_PIXELS / 2), BOARD_Y + (BOARD_HEIGHT_PIXELS / 2), msg, 8);
      g.drawString(msg, BOARD_X + (BOARD_WIDTH_PIXELS / 2) - (font.getWidth(msg) / 2), BOARD_Y + (BOARD_HEIGHT_PIXELS / 2) - (font.getHeight(msg) / 2));
    }

    // If game over, draw winner

    if (state == 2) {
      Player winner = null;
      for (Map.Entry<Player, Snake> entry : currentGamePlayers.entrySet()) {
        Player player = entry.getKey();
        Snake snake = entry.getValue();
        if (!snake.dead) {
          winner = player;
        }
      }

      AngelCodeFont font = resources.getFont("FONT_NOKIA48");
      int x = BOARD_X + (BOARD_WIDTH_PIXELS / 2);
      int y = BOARD_Y + (BOARD_HEIGHT_PIXELS / 2) - font.getLineHeight();

      g.setColor(Color.blue);

      if (winner != null) {
        SlickGraphicUtil.drawTextCentered(g, font, x, y, "WINNER!");
        y += font.getLineHeight();
        SlickGraphicUtil.drawTextCentered(g, font, x, y, winner.getName());
      }
      else {
        SlickGraphicUtil.drawTextCentered(g, font, x, y, "YOU'RE ALL");
        y += font.getLineHeight();
        SlickGraphicUtil.drawTextCentered(g, font, x, y, "LOSERS!");
      }
    }

    // GRID
//    drawGrid(g);

//    Image screenglare = resources.getImage("IMAGE_SCREENGLARE");
//    g.drawImage(screenglare, BOARD_X, BOARD_Y,
//            BOARD_X + BOARD_WIDTH_PIXELS, BOARD_Y + BOARD_HEIGHT_PIXELS,
//            0, 0, screenglare.getWidth(), screenglare.getHeight());

    // SIDEBAR

    drawScoreBoard(g);

    // Queue for next game

    if (SnaxeGame.DEBUG) {
      g.drawString("Speed:" + timeBetweenMoves, 10, 500);
    }

    drawQueue(g, (SnaxeGame) game);
  }

  private void drawScoreBoard(Graphics g)
  {
    int y = BOARD_Y;

    AngelCodeFont font = resources.getFont("FONT_NOKIA24");
    g.setFont(font);

    for (Map.Entry<Player, Snake> entry : currentGamePlayers.entrySet()) {
      Player player = entry.getKey();
      Snake snake = entry.getValue();

      Color color;
      if (snake.dead) color = new Color(snake.color.r, snake.color.g, snake.color.b, 0.8f);//DEAD_COLOR;
      else color = snake.color;

      g.setColor(color);
      g.drawString(player.getName(), 10, y);
      y += font.getLineHeight();

      int x = 10;
      if (snake.dead) {
        Image im = resources.getImage("IMAGE_SCOREBOARD_DEAD");
        g.setColor(Color.white);
        g.drawImage(im, x, y + 3);
        x += im.getWidth() + 4;
        g.setColor(color);
      }
      g.drawString("" + snake.score, x, y);
      y += font.getLineHeight() * 1.5;
    }
  }

  private void drawQueue(Graphics g, SnaxeGame game)
  {
    AngelCodeFont font = resources.getFont("FONT_NOKIA16");
    g.setFont(font);

    int numQueued = 0;
    for (Player player : game.getSortedPlayers()) {
      if (currentGamePlayers.containsKey(player)) continue;
      if (player.getQueueJoinTime() <= 0) continue;
      numQueued++;
    }

    int y;
    if (numQueued == 0) {
      y = QUEUE_BOTTOM_Y - (2 * font.getLineHeight());
    }
    else {
      y = QUEUE_BOTTOM_Y - ((numQueued + 4) * font.getLineHeight());
    }

    g.setColor(Color.white);
    g.drawString("Connect to wi-fi \"Snaxe\"", QUEUE_LEFT_X, y);
    y += font.getLineHeight();
    g.drawString("and visit http://game/", QUEUE_LEFT_X, y);
    y += font.getLineHeight();


    if (numQueued == 0) return;
    y += font.getLineHeight();

    g.setColor(Color.gray);

    g.drawString("Queued for next game:", QUEUE_LEFT_X, y);
    y += font.getLineHeight();

    for (Player player : game.getSortedPlayers()) {
      if (currentGamePlayers.containsKey(player)) continue;
      if (player.getQueueJoinTime() <= 0) continue;

      g.drawString(player.getName(), QUEUE_LEFT_X, y);
      y += font.getLineHeight();
    }
  }

  private void drawGrid(Graphics g)
  {
    g.setColor(Color.gray);
    for (int x = 0; x <= BOARD_WIDTH; ++x) {
      g.drawLine(BOARD_X + (x * SQUARE_WIDTH), BOARD_Y, BOARD_X + (x * SQUARE_WIDTH), BOARD_Y + (BOARD_HEIGHT * SQUARE_HEIGHT));
    }
    for (int y = 0; y <= BOARD_HEIGHT; ++y) {
      g.drawLine(BOARD_X, BOARD_Y + (y * SQUARE_HEIGHT), BOARD_X + BOARD_WIDTH_PIXELS, BOARD_Y + (y * SQUARE_HEIGHT));
    }
  }

  private void drawStringScaled(Graphics g, int x, int y, String text, float scale)
  {
//    SGL GL = Renderer.get();
//    GL.glDisable(SGL.GL_BLEND);
    g.pushTransform();
    g.translate(x, y);
    g.scale(scale, scale);
    g.drawString(text, 0, 0);
    g.popTransform();
//    g.setAntiAlias(true);
  }

  private Image getImageForPart(SnakePart part)
  {
    Image im = null;
    if (part.previous == null) {
      //head
      if (part.next == null) {

      }
      else {
        switch (part.nextPieceDir) {
          case NORTH:
            im = resources.getImage("IMAGE_SNAKE_HEAD_S");
            break;
          case EAST:
            im = resources.getImage("IMAGE_SNAKE_HEAD_W");
            break;
          case SOUTH:
            im = resources.getImage("IMAGE_SNAKE_HEAD_N");
            break;
          case WEST:
            im = resources.getImage("IMAGE_SNAKE_HEAD_E");
            break;
        }
      }
    }
    else if (part.next == null) {
      //tail
      switch (part.prevPieceDir) {
        case NORTH:
          im = resources.getImage("IMAGE_SNAKE_TAIL_N");
          break;
        case EAST:
          im = resources.getImage("IMAGE_SNAKE_TAIL_E");
          break;
        case SOUTH:
          im = resources.getImage("IMAGE_SNAKE_TAIL_S");
          break;
        case WEST:
          im = resources.getImage("IMAGE_SNAKE_TAIL_W");
          break;
      }
    }
    else {
      switch (part.prevPieceDir) {
        case NORTH:
          switch (part.nextPieceDir) {
            case EAST:
              im = resources.getImage("IMAGE_SNAKE_BODY_NE");
              break;
            case SOUTH:
              im = resources.getImage("IMAGE_SNAKE_BODY_V");
              break;
            case WEST:
              im = resources.getImage("IMAGE_SNAKE_BODY_NW");
              break;
          }
          break;
        case EAST:
          switch (part.nextPieceDir) {
            case NORTH:
              im = resources.getImage("IMAGE_SNAKE_BODY_NE");
              break;
            case SOUTH:
              im = resources.getImage("IMAGE_SNAKE_BODY_SE");
              break;
            case WEST:
              im = resources.getImage("IMAGE_SNAKE_BODY_H");
              break;
          }
          break;
        case SOUTH:
          switch (part.nextPieceDir) {
            case NORTH:
              im = resources.getImage("IMAGE_SNAKE_BODY_V");
              break;
            case EAST:
              im = resources.getImage("IMAGE_SNAKE_BODY_SE");
              break;
            case WEST:
              im = resources.getImage("IMAGE_SNAKE_BODY_SW");
              break;
          }
          break;
        case WEST:
          switch (part.nextPieceDir) {
            case NORTH:
              im = resources.getImage("IMAGE_SNAKE_BODY_NW");
              break;
            case EAST:
              im = resources.getImage("IMAGE_SNAKE_BODY_H");
              break;
            case SOUTH:
              im = resources.getImage("IMAGE_SNAKE_BODY_SW");
              break;
          }
          break;
      }
    }
    return im;
  }

  @Override
  public void keyPressed(int key, char c)
  {
  }

  public void addPowerUpRiseEvent(int squareX, int squareY, String text)
  {
    int x = BOARD_X + (squareX * SQUARE_WIDTH) + (SQUARE_WIDTH / 2);
    int y = BOARD_Y + (squareY * SQUARE_HEIGHT);

    x -= powerUpFont.getWidth(text) / 2;
    y -= powerUpFont.getLineHeight() / 2;

    eventManager.addEvent(new PowerUpRiseEvent(curTime, curTime + 1500, x, y, text));
  }

  private void addPlayer(Player player, int playerNum, int numPlayers)
  {
    int x = START_POSITIONS[numPlayers - 2][playerNum].x;
    int y = START_POSITIONS[numPlayers - 2][playerNum].y;
    Direction facing = START_FACING[numPlayers - 2][playerNum];

    Snake snake = new Snake(x, y, facing);

    snake.color = SNAKE_COLORS[playerNum];

    try {
      player.sendMessage(Protocol.MESSAGE_OUT_PLAY);
      player.sendMessage(Protocol.MESSAGE_OUT_COLOR + ColorUtilSlick.colorToHexRGB(snake.color));
    }
    catch (IOException e) {
      log.log(Level.SEVERE, "Unable to send color to client " + player);
      return;
    }

    // Move them one space in that direction
    snake.move();

    board.snakes.add(snake);

    currentGamePlayers.put(player, snake);

    // Remove them from the queue
    player.setQueueJoinTime(0);
  }


  // Player quit mid-game. TODO should show them on the scoreboard still, but crossed out
  public void removePlayer(Player player)
  {
    Snake snake = currentGamePlayers.get(player);

    if (snake == null) {
      log.log(Level.WARNING, "Quit for player with no snake");
    }

    currentGamePlayers.remove(player);
    board.snakes.remove(snake);
  }

  public void playerAction(Player player, String message)
  {
    Snake snake = currentGamePlayers.get(player);

    if (snake == null) {
      log.log(Level.WARNING, "Message for player with no snake: " + message);
      return;
    }

    if (Protocol.MESSAGE_IN_NORTH.equals(message)) {
      snake.setFacing(Direction.NORTH);
    }
    else if (Protocol.MESSAGE_IN_EAST.equals(message)) {
      snake.setFacing(Direction.EAST);
    }
    else if (Protocol.MESSAGE_IN_SOUTH.equals(message)) {
      snake.setFacing(Direction.SOUTH);
    }
    else if (Protocol.MESSAGE_IN_WEST.equals(message)) {
      snake.setFacing(Direction.WEST);
    }
    else {
      log.log(Level.WARNING, "Unrecognised message " + message + " from player " + player);
    }
  }
}
