package net.davidc.egp.snaxe;

import javax.swing.*;
import java.io.IOException;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.goofans.gootool.util.DebugUtil;
import net.davidc.egp.common.slick2d.JavaUtilLogSystem;
import net.davidc.egp.common.resources.ResourcesException;
import net.davidc.egp.common.slick2d.resources.SlickResources;
import net.davidc.egp.snaxe.networking.SwarmListener;
import net.davidc.egp.snaxe.networking.Protocol;
import net.davidc.egp.snaxe.networking.SwarmScreenClient;
import net.davidc.egp.snaxe.states.InsertCoinState;
import net.davidc.egp.snaxe.states.PlayState;
import org.newdawn.slick.AppGameContainer;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Input;
import org.newdawn.slick.SlickException;
import org.newdawn.slick.state.StateBasedGame;

/**
 * @author David Croft (david@davidc.net)
 * @version $Id: SnaxeGame.java 25 2011-09-18 15:41:54Z david $
 */
public class SnaxeGame extends StateBasedGame implements SwarmListener
{
  private static final Logger log = Logger.getLogger(SnaxeGame.class.getName());

  public static final int MIN_PLAYERS = 2;
  public static final int MAX_PLAYERS = 8; // Cannot exceed 8, the length of position/colour arrays

  public static final boolean DEBUG = false;

  private static final String NAME_LEFT_HAND = "Player1";
  private static final String NAME_RIGHT_HAND = "Player2";
  private static final String NAME_THIRD_HAND = "Player3";
  private static final String NAME_FOURTH_HAND = "Player4";

  private SlickResources globalResources;
  private SwarmScreenClient swarmClient;

  private Set<Player> players;

  private long masterTime;
  private GameContainer container;

  public SnaxeGame() throws ResourcesException
  {
    super("Snaxe");

    try {
      swarmClient = new SwarmScreenClient();
      swarmClient.connect("localhost", 8002);
      swarmClient.addSwarmListener(this);

      // Broadcast "gameover" to reinitialise all clients
      swarmClient.broadcastMessage(Protocol.MESSAGE_OUT_GAMEOVER);
    }
    catch (IOException e) {
      log.log(Level.WARNING, "Communication error with swarm server", e);
      JOptionPane.showMessageDialog(null, "Snaxe is unable to communicate with the swarm server.\n" +
              "Network play will not be available.\n\n" + e.getClass().getName() + "\n" + e.getMessage(),
              "Local play only", JOptionPane.WARNING_MESSAGE);
      swarmClient = null;
    }

    players = new HashSet<Player>();
  }

  @Override
  public void initStatesList(GameContainer container) throws SlickException
  {
    this.container = container;
    globalResources = new SlickResources();
    try {
      globalResources.loadFromClasspath("/resources.xml", "global");
    }
    catch (ResourcesException e) {
      throw new SlickException("Unable to load resources", e);
    }

//    addState(new LoadingState());
    addState(new InsertCoinState());
    addState(new PlayState());
  }

  @Override
  public void keyPressed(int key, char c)
  {
    if (key == Input.KEY_F12 && getCurrentState() instanceof InsertCoinState) {
      ((InsertCoinState) getCurrentState()).forceStart();
      return;
    }

    // Left hand
    if (handleLocalPlayerKeyPress(LocalPlayer.PLAYER_LOCAL_LEFT_HAND, NAME_LEFT_HAND, key,
            Input.KEY_Q, Input.KEY_E, Input.KEY_W, Input.KEY_A, Input.KEY_S, Input.KEY_D))
      return;

    // Right hand
    if (handleLocalPlayerKeyPress(LocalPlayer.PLAYER_LOCAL_RIGHT_HAND, NAME_RIGHT_HAND, key,
            Input.KEY_U, Input.KEY_O, Input.KEY_I, Input.KEY_J, Input.KEY_K, Input.KEY_L))
      return;

    // Arrow hand
    if (handleLocalPlayerKeyPress(LocalPlayer.PLAYER_LOCAL_ARROW_HAND, NAME_THIRD_HAND, key,
            Input.KEY_DELETE, Input.KEY_END, Input.KEY_UP, Input.KEY_LEFT, Input.KEY_DOWN, Input.KEY_RIGHT))
      return;

    // Numpad hand
    if (handleLocalPlayerKeyPress(LocalPlayer.PLAYER_LOCAL_NUMPAD_HAND, NAME_FOURTH_HAND, key,
            Input.KEY_NUMPAD7, Input.KEY_NUMPAD9, Input.KEY_NUMPAD8, Input.KEY_NUMPAD4, Input.KEY_NUMPAD5, Input.KEY_NUMPAD6))
      return;

    if (key == Input.KEY_RETURN && container.getInput().isKeyDown(Input.KEY_LALT)) {
      try {
        container.setFullscreen(!container.isFullscreen());
        container.setMouseGrabbed(container.isFullscreen());
      }
      catch (SlickException e) {
        log.log(Level.WARNING, "Can't toggle full-screen", e);
      }
      return;
    }

    super.keyPressed(key, c);
  }

  private boolean handleLocalPlayerKeyPress(int hand, String name, int keyPressed, int keyJoin, int keyQuit, int keyNorth, int keyWest, int keySouth, int keyEast)
  {
    Player player = getLocalPlayer(hand);

    if (player == null && keyPressed == keyJoin) {
      addLocalPlayer(hand, name);
      return true;
    }
    else if (player != null) {
      if (keyPressed == keyNorth) {
        handlePlayerMessage(player, Protocol.MESSAGE_IN_NORTH);
        return true;
      }
      else if (keyPressed == keyWest) {
        handlePlayerMessage(player, Protocol.MESSAGE_IN_WEST);
        return true;
      }
      else if (keyPressed == keySouth) {
        handlePlayerMessage(player, Protocol.MESSAGE_IN_SOUTH);
        return true;
      }
      else if (keyPressed == keyEast) {
        handlePlayerMessage(player, Protocol.MESSAGE_IN_EAST);
        return true;
      }
      else if (keyPressed == keyJoin) {
        handlePlayerMessage(player, Protocol.MESSAGE_IN_NAME + Protocol.SEPARATOR + name);
        return true;
      }
      else if (keyPressed == keyQuit) {
        handlePlayerQuit(player);
        return true;
      }
    }
    return false;
  }

  private void addLocalPlayer(int hand, String name)
  {
    Player player = new LocalPlayer(hand);
    player.setName(name);
    handleNewPlayer(player);
    handlePlayerMessage(player, Protocol.MESSAGE_IN_NAME + Protocol.SEPARATOR + name);
  }

  @Override
  protected void preUpdateState(GameContainer container, int delta) throws SlickException
  {
    masterTime += delta;

    if (swarmClient != null) {
      try {
        swarmClient.poll();
      }
      catch (IOException e) {
        log.log(Level.WARNING, "Error polling network", e);
        throw new SlickException("Error polling network", e);
      }
    }
  }

  public void playerJoined(int playerId)
  {
    Player player = new SwarmPlayer(swarmClient, playerId);
    handleNewPlayer(player);
  }

  private void handleNewPlayer(Player player)
  {
    log.log(Level.FINER, "Player " + player + " joined");
    players.add(player);
  }

  public void messageReceived(int playerId, String message)
  {
    Player player = getSwarmPlayer(playerId);
    if (player == null) {
      log.log(Level.WARNING, "Message from unrecognised player " + playerId + ":  " + message);
      return;
    }
    handlePlayerMessage(player, message);
  }

  private void handlePlayerMessage(Player player, String message)
  {
    log.log(Level.FINER, "Message from player " + player + ": " + message);

    String[] bits = message.split(Protocol.SEPARATOR, 2);

    if (Protocol.MESSAGE_IN_NAME.equals(bits[0])) {
      player.setName(bits[1]);
      player.setQueueJoinTime(masterTime);
      try {
        player.sendMessage(Protocol.MESSAGE_OUT_QUEUED);
      }
      catch (IOException e) {
        log.log(Level.SEVERE, "Unable to send queued message to " + player);
      }
      return;
    }

    if (getCurrentState() instanceof PlayState) {
      ((PlayState) getCurrentState()).playerAction(player, message);
    }
  }

  public void playerQuit(int playerId)
  {
    Player player = getSwarmPlayer(playerId);
    if (player == null) {
      log.log(Level.WARNING, "Quit from unrecognised player " + playerId);
      return;
    }
    handlePlayerQuit(player);
  }

  private void handlePlayerQuit(Player player)
  {
    log.log(Level.FINER, "Player " + player + " quit");
    players.remove(player);

    if (getCurrentState() instanceof PlayState) {
      ((PlayState) getCurrentState()).removePlayer(player);
    }
  }

  public SlickResources getGlobalResources()
  {
    return globalResources;
  }

  private Player getLocalPlayer(int hand)
  {
    for (Player player : players) {
      if (player instanceof LocalPlayer && ((LocalPlayer) player).getHand() == hand) {
        return player;
      }
    }
    return null;
  }

  private Player getSwarmPlayer(int playerId)
  {
    for (Player player : players) {
      if (player instanceof SwarmPlayer && ((SwarmPlayer) player).getPlayerId() == playerId) {
        return player;
      }
    }
    return null;
  }

  public List<Player> getSortedPlayers()
  {
    List<Player> sortedPlayers = new ArrayList<Player>(players);
    Collections.sort(sortedPlayers, new Player.PlayerQueueJoinTimeComparator());
    return sortedPlayers;
  }

  public static void main(String[] args)
  {
    try {
      UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());

      DebugUtil.setAllLogging();
      DebugUtil.setTerseLogFormatter();
      JavaUtilLogSystem.installLogSystem();

      AppGameContainer app = new AppGameContainer(new SnaxeGame());

      app.setDisplayMode(1024, 768, false);
      app.setShowFPS(false);
      app.setTargetFrameRate(60);
      app.start();
    }
    catch (Throwable t) {
      log.log(Level.SEVERE, "Throwable reached main method", t);
      JOptionPane.showMessageDialog(null, "Exception in main thread:\n" + t.getClass().getName() + "\n" + t.getMessage(),
              "SNAXE THROWN", JOptionPane.ERROR_MESSAGE);
    }
  }
}
