package net.davidc.egp.snaxe.board;

import java.awt.Point;
import java.util.*;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.davidc.egp.snaxe.states.PlayState;

/**
 * @author David Croft (david@davidc.net)
 * @version $Id: GameBoard.java 26 2011-09-18 16:24:33Z david $
 */
public class GameBoard
{
  private static final Logger log = Logger.getLogger(GameBoard.class.getName());

  private static final int RABBIT_SCORE = 20;

  private static final float RABBIT_SPAWN_CHANCE = 0.2f;
  private static final int RABBIT_PIECE_MIN_LIFETIME_MOVES = 15;
  private static final int RABBIT_PIECE_MAX_LIFETIME_MOVES = 45;

  private static final float INVULNERABILITY_SPAWN_CHANCE = 0.01f;
  private static final int INVULNERABILITY_PIECE_MIN_LIFETIME_MOVES = 15;
  private static final int INVULNERABILITY_PIECE_MAX_LIFETIME_MOVES = 30;
  private static final int INVULNERABILITY_DURATION_TIME = 20 * 1000;

  private static final float SPEED_SPAWN_CHANCE = 0.01f;
  private static final int SPEED_PIECE_MIN_LIFETIME_MOVES = 15;
  private static final int SPEED_PIECE_MAX_LIFETIME_MOVES = 25;
  private static final int SPEED_DURATION_MOVES = 15;

  private static final float SKULL_SPAWN_CHANCE = 0.05f;
  private static final int SKULL_PIECE_MIN_LIFETIME_MOVES = 5;
  private static final int SKULL_PIECE_MAX_LIFETIME_MOVES = 20;
  private static final int SKULL_DEAD_SNAKE_LIFETIME_MOVES = 15;

private  int width;
private  int height;
  private PlayState playState;
  protected Set<Snake> snakes;
  protected Set<Piece> pieces;

  private final Random random = new Random();

  private int minRabbits;
  private int maxRabbits;

  private  long lastTime = 0; // last time move() was called

  public GameBoard(int width, int height, PlayState playState)
  {
    this.width = width;
    this.height = height;
    this.playState = playState;
    snakes = new HashSet<Snake>();
    pieces = new HashSet<Piece>();
  }

  public void initRabbits()
  {
    minRabbits = (int) Math.ceil(snakes.size() / 4.0);
    maxRabbits = (int) Math.ceil(snakes.size() * 0.75);

    populateRabbits();
//    addPiece(PieceType.SPEED, 10, 10, 1000, 1000); // testing - static piece
  }

  public void move(long curTime)
  {
    int diffTime = (int) (curTime - lastTime);
    lastTime = curTime;

    // Move all snakes
    for (Snake snake : snakes) {
      moveSnake(snake);

      // Reduce invulnerability
      if (snake.isInvulnerable()) {
        int newRemaining = snake.getInvulnerabilityTimeRemaining() - diffTime;
        if (newRemaining < 0) newRemaining = 0;
        snake.setInvulnerabilityTimeRemaining(newRemaining);

        if (newRemaining == 0) {
          invulnerabilityEnded(snake);
        }
      }

      if (snake.isSpeedy()) {
        snake.setSpeedMovesRemaining(snake.getSpeedMovesRemaining() - 1);
      }
    }

    // if a snake is now off-grid, it dies.
    for (Snake snake : snakes) {
      checkSnakeOffGrid(snake);
    }

    // if two snakes occupy the same position, both die
    for (Snake snake : snakes) {
      checkSnakeHeadbuttSnake(snake);
    }

    // if a snake occupies the position of someone else's body or tail, they die
    for (Snake snake : snakes) {
      SnakePart partToCheck = snake.getHead();
      checkSnakeOnSnake(snake, partToCheck);
    }

    // if a snake occupies the position of a board piece, they die or gain the power-up
    for (Snake snake : snakes) {
      SnakePart partToCheck = snake.getHead();
      checkSnakeOnPiece(snake, partToCheck);
    }

    // Kill all dying snakes
    killDeadSnakes();

    // if a piece occupies the space of another piece (e.g. two skulls, or a skull on top of a rabbit), keep
    // only one (currently random-ish)
    killStackedPieces();

    // Now decay old pieces
    decayPieces();

    // Now add new pieces
    populateRabbits();
    spawnSpecials();
  }

  // Half-move, for speedy snakes only
  public void halfMove()
  {
    // Move all snakes
    for (Snake snake : snakes) {
      if (snake.isSpeedy()) {
        moveSnake(snake);
      }
    }

    // if a snake is now off-grid, it dies.
    for (Snake snake : snakes) {
      if (snake.isSpeedy()) {
        checkSnakeOffGrid(snake);

        // if two snakes occupy the same position, both die
        checkSnakeHeadbuttSnake(snake);

        // if a snake occupies the position of someone else's body or tail, they die
        SnakePart partToCheck = snake.getHead();
        checkSnakeOnSnake(snake, partToCheck);

        // if a snake occupies the position of a board piece, they die or gain the power-up
        checkSnakeOnPiece(snake, partToCheck);
      }
    }

    // Kill all dying snakes
    killDeadSnakes();
  }

  private void moveSnake(Snake snake)
  {
    snake.move();
  }

  private void checkSnakeOffGrid(Snake snake)
  {
    SnakePart head = snake.getHead();
    if (head.getX() < 0 || head.getX() >= width || head.getY() < 0 || head.getY() >= height) {
      log.log(Level.FINE, "Snake " + snake + " went off-board and died.");
      snake.setDead(true);
    }
  }

  private void checkSnakeHeadbuttSnake(Snake snake1)
  {
    if (snake1.isDead() || snake1.isInvulnerable()) return;

    for (Snake snake2 : snakes) {
      if (snake1 == snake2) continue;
      if (snake2.isInvulnerable()) continue;

      SnakePart head1 = snake1.getHead();
      SnakePart head2 = snake2.getHead();
      if (head1.getX() == head2.getX() && head1.getY() == head2.getY()) {
        log.log(Level.FINE, "Snake " + snake1 + " headbutted " + snake2 + ". Both died.");
        snake1.setDead(true);
        snake2.setDead(true);
      }
    }
  }

  private void checkSnakeOnSnake(Snake snake1, SnakePart partToCheck)
  {
    if (snake1.isDead() || snake1.isInvulnerable()) return;

    int x = partToCheck.getX();
    int y = partToCheck.getY();

    nextSnake2:
    for (Snake snake2 : snakes) {
      if (snake2.isInvulnerable()) continue;

      SnakePart part = snake2.getHead();
//        if (snake1 == snake2) part = part.next; // allow to touch their own head! - now handled by partToCheck != part below
      while (part != null) {
        if (partToCheck != part && x == part.getX() && y == part.getY()) {
          log.log(Level.FINE, "Snake " + snake1 + " hit the tail of " + snake2 + " and died.");
          snake1.setDead(true);
          break nextSnake2;
        }
        part = part.getNext();
      }
    }
  }

  private void checkSnakeOnPiece(Snake snake, SnakePart partToCheck)
  {
    if (snake.isDead()) return;

    int x = partToCheck.getX();
    int y = partToCheck.getY();

    nextPiece:
    for (Piece piece : pieces) {
      if (x == piece.getX() && y == piece.getY()) {

        switch (piece.getType()) {
          case RABBIT:
            snake.addScore(RABBIT_SCORE);
            snake.grow(1);
            playState.addPowerUpRiseEvent(piece.getX(), piece.getY(), "+" + RABBIT_SCORE);
            log.log(Level.FINE, "Snake " + snake + " ate a rabbit, score now " + snake.getScore());
            break;
          case SKULL:
            if (snake.isInvulnerable()) continue nextPiece;
            log.log(Level.FINE, "Snake " + snake + " was killed by a skull.");
            snake.setDead(true);
            break;
          case INVULNERABILITY:
            snake.setInvulnerabilityTimeRemaining(INVULNERABILITY_DURATION_TIME);
            playState.addPowerUpRiseEvent(piece.getX(), piece.getY(), "INVULNERABLE!");
            log.log(Level.FINE, "Snake " + snake + " got invulnerability");
            break;
          case SPEED:
            snake.setSpeedMovesRemaining(SPEED_DURATION_MOVES);
            playState.addPowerUpRiseEvent(piece.getX(), piece.getY(), "+SPEED");
            log.log(Level.FINE, "Snake " + snake + " got speed");
            break;
        }

        pieces.remove(piece);
        return; // should never be two pieces on top of each other, so this is ok
      }
    }
  }

  private void killDeadSnakes()
  {
    boolean done = false;
    killSnakes:
    while (!done) {
      for (Snake snake : snakes) {
        if (snake.isDead()) {
          // Replace all pieces by skulls
          log.log(Level.FINE, "Snake " + snake + " is replaced by skulls.");
          SnakePart part = snake.getHead().getNext(); // don't place a skull on the head itself, cos that's the position of whatever killed them!
          while (part != null) {
            addPiece(PieceType.SKULL, part.getX(), part.getY(), SKULL_DEAD_SNAKE_LIFETIME_MOVES, SKULL_DEAD_SNAKE_LIFETIME_MOVES);
            part = part.getNext();
          }

          // Remove snake
          snakes.remove(snake);
          // Modified our iterators; must restart
          continue killSnakes;
        }
      }
      done = true;
    }
  }

  private void killStackedPieces()
  {
    boolean done;

    done = false;
    restartPiece:
    while (!done) {
      for (Piece piece1 : pieces) {
        for (Piece piece2 : pieces) {
          if (piece1 == piece2) continue;
          if (piece1.getX() == piece2.getX() && piece1.getY() == piece2.getY()) {
            log.log(Level.FINE, "Piece " + piece1 + " is on top of " + piece2 + ", the latter is removed.");
            pieces.remove(piece2);
            // Modified our iterators; must restart
            continue restartPiece;
          }
        }
      }
      done = true;
    }
  }

  private void decayPieces()
  {
    List<Piece> piecesToRemove = new ArrayList<Piece>();

    for (Piece piece : pieces) {
      int newLifetime = piece.getLifetimeRemaining() - 1;
      if (newLifetime < 0) newLifetime = 0;
      piece.setLifetimeRemaining(newLifetime);

      if (newLifetime == 0) {
        piecesToRemove.add(piece);
        log.log(Level.FINE, "Piece " + piece + " decayed");
      }
    }

    for (Piece piece : piecesToRemove) {
      pieces.remove(piece);
    }
  }

  private void invulnerabilityEnded(Snake snake)
  {
    //Snake just exited invulnerability. Check that none of its parts cross each other, another snake, or a skull

    // if two snakes occupy the same position, both die
    checkSnakeHeadbuttSnake(snake);

    SnakePart part = snake.getHead();
    while (part != null) {
      // if this part occupies the position of someone else's or their own body or tail, they die
      checkSnakeOnSnake(snake, part);

      // if this part occupies the position of a board piece, they die or gain the power-up
      checkSnakeOnPiece(snake, part);

      part = part.getNext();
    }
  }

  private void populateRabbits()
  {
    int numRabbits = countPieces(PieceType.RABBIT);
    if (numRabbits >= maxRabbits) return;

    while (numRabbits < minRabbits) {
      addRandomRabbit();
      numRabbits++;
    }

    // Chance of adding an extra rabbit

    if (random.nextFloat() < RABBIT_SPAWN_CHANCE) {
      addRandomRabbit();
    }
  }

  private void spawnSpecials()
  {
    int numSpecials = countPieces(PieceType.INVULNERABILITY) + countPieces(PieceType.SPEED);
    if (numSpecials == 0) {
      if (random.nextFloat() < INVULNERABILITY_SPAWN_CHANCE) {
        addRandomInvulnerability();
      }
      else if (random.nextFloat() < SPEED_SPAWN_CHANCE) {
        addRandomSpeed();
      }
    }

    if (random.nextFloat() < SKULL_SPAWN_CHANCE) {
      addRandomSkull();
    }
  }

  private int countPieces(PieceType type)
  {
    int numPiecesOfType = 0;
    for (Piece piece : pieces) {
      if (piece.getType()== type) {
        numPiecesOfType++;
      }
    }
    return numPiecesOfType;
  }

  private void addRandomRabbit()
  {
    Point p = getRandomUnoccupiedPoint();
    if (p == null) return;

    log.log(Level.FINE, "Add rabbit at " + p.x + "," + p.y);
    addPiece(PieceType.RABBIT, p.x, p.y, RABBIT_PIECE_MIN_LIFETIME_MOVES, RABBIT_PIECE_MAX_LIFETIME_MOVES);
  }

  private void addRandomInvulnerability()
  {
    Point p = getRandomUnoccupiedPoint();
    if (p == null) return;

    log.log(Level.FINE, "Add invulnerability at " + p.x + "," + p.y);
    addPiece(PieceType.INVULNERABILITY, p.x, p.y, INVULNERABILITY_PIECE_MIN_LIFETIME_MOVES, INVULNERABILITY_PIECE_MAX_LIFETIME_MOVES);
  }

  private void addRandomSpeed()
  {
    Point p = getRandomUnoccupiedPoint();
    if (p == null) return;

    log.log(Level.FINE, "Add speed at " + p.x + "," + p.y);
    addPiece(PieceType.SPEED, p.x, p.y, SPEED_PIECE_MIN_LIFETIME_MOVES, SPEED_PIECE_MAX_LIFETIME_MOVES);
  }

  private void addRandomSkull()
  {
    Point p = getRandomUnoccupiedPoint();
    if (p == null) return;

    log.log(Level.FINE, "Add skull at " + p.x + "," + p.y);
    addPiece(PieceType.SKULL, p.x, p.y, SKULL_PIECE_MIN_LIFETIME_MOVES, SKULL_PIECE_MAX_LIFETIME_MOVES);
  }

  /**
   * Find a point that isn't occupied by any snake part or piece.
   *
   * @return an unoccupied point, or null if none could be found.
   */
  private Point getRandomUnoccupiedPoint()
  {
    Point p = new Point();

    newPoint:
    for (int i = 0; i < 50; ++i) {
      p.x = random.nextInt(width);
      p.y = random.nextInt(height);

      // Make sure there's not an existing piece or snake part there
      for (Piece piece : pieces) {
        if (piece.getX() == p.x && piece.getY() == p.y) continue newPoint;
      }

      for (Snake snake : snakes) {
        SnakePart part = snake.getHead();
        while (part != null) {
          if (part.getX() == p.x && part.getY() == p.y) continue newPoint;
          part = part.getNext();
        }
      }

      return p;
    }

    log.log(Level.WARNING, "Couldn't find an unoccupied point fast enough");
    return null;
  }

  private void addPiece(PieceType type, int x, int y, int minLifetime, int maxLifetime)
  {
    int lifetime = minLifetime + random.nextInt(maxLifetime + 1 - minLifetime);
    pieces.add(new Piece(type, x, y, lifetime));
  }

  public Set<Snake> getSnakes()
  {
    return snakes;
  }

  public void addSnake(Snake snake)
  {
    snakes.add(snake);
  }

  public void removeSnake(Snake snake)
  {
    snakes.remove(snake);
  }

  public Set<Piece> getPieces()
  {
    return pieces;
  }
}
